@@ -34,6 +34,7 @@ from orca import debug
 from orca import messages
 from orca import object_properties
 from orca import orca_state
+from orca import settings
 from orca import settings_manager
 from orca import speech_generator
 
@@ -45,10 +46,33 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
     def __init__(self, script):
         super().__init__(script)
 
+    def _generateOldAncestors(self, obj, **args):
+        if args.get('index', 0) > 0:
+            return []
+
+        priorObj = args.get('priorObj')
+        if self._script.utilities.isInlineIframeDescendant(priorObj):
+            return []
+
+        return super()._generateOldAncestors(obj, **args)
+
+    def _generateNewAncestors(self, obj, **args):
+        if args.get('index', 0) > 0 \
+           and not self._script.utilities.isListDescendant(obj):
+            return []
+
+        if self._script.utilities.isInlineIframeDescendant(obj):
+            return []
+
+        return super()._generateNewAncestors(obj, **args)
+
     def _generateAncestors(self, obj, **args):
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateAncestors(obj, **args)
 
+        if self._script.inSayAll() and obj == orca_state.locusOfFocus:
+            return []
+
         result = []
         priorObj = args.get('priorObj')
         if priorObj and self._script.utilities.inDocumentContent(priorObj):
@@ -70,11 +94,15 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                                pyatspi.ROLE_MENU_BAR,
                                pyatspi.ROLE_TOOL_BAR]
         args['skipRoles'] = [pyatspi.ROLE_PARAGRAPH,
+                             pyatspi.ROLE_HEADING,
                              pyatspi.ROLE_LABEL,
                              pyatspi.ROLE_LINK,
                              pyatspi.ROLE_LIST_ITEM,
                              pyatspi.ROLE_TEXT]
 
+        if self._script.utilities.isEditableDescendantOfComboBox(obj):
+            args['skipRoles'].append(pyatspi.ROLE_COMBO_BOX)
+
         result.extend(super()._generateAncestors(obj, **args))
 
         return result
@@ -97,7 +125,35 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         # call utility methods rather than generate it.
         return super()._generateAnyTextSelection(obj, **args)
 
+    def _generateHasPopup(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        if not self._script.utilities.inDocumentContent(obj):
+            return []
+
+        result = []
+        popupType = self._script.utilities.popupType(obj)
+        if popupType == 'dialog':
+            result = [messages.HAS_POPUP_DIALOG]
+        elif popupType == 'grid':
+            result = [messages.HAS_POPUP_GRID]
+        elif popupType == 'listbox':
+            result = [messages.HAS_POPUP_LISTBOX]
+        elif popupType in ('menu', 'true'):
+            result = [messages.HAS_POPUP_MENU]
+        elif popupType == 'tree':
+            result = [messages.HAS_POPUP_TREE]
+
+        if result:
+            result.extend(self.voice(speech_generator.SYSTEM))
+
+        return result
+
     def _generateClickable(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return []
 
@@ -113,15 +169,29 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return []
 
     def _generateDescription(self, obj, **args):
-        if self._script.utilities.isZombie(obj) \
-           or obj != orca_state.locusOfFocus:
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateDescription(obj, **args)
+
+        if self._script.utilities.isZombie(obj):
+            return []
+
+        if self._script.utilities.preferDescriptionOverName(obj):
+            return []
+
+        role = args.get('role', obj.getRole())
+        if obj != orca_state.locusOfFocus:
+            if role in [pyatspi.ROLE_ALERT, pyatspi.ROLE_DIALOG]:
+                return super()._generateDescription(obj, **args)
+            if not args.get('inMouseReview'):
+                return []
+
         formatType = args.get('formatType')
         if formatType == 'basicWhereAmI' and self._script.utilities.isLiveRegion(obj):
             return self._script.liveRegionManager.generateLiveRegionDescription(obj, **args)
 
-        role = args.get('role', obj.getRole())
         if role == pyatspi.ROLE_TEXT and formatType != 'basicWhereAmI':
             return []
 
@@ -132,6 +202,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateDescription(obj, **args)
 
     def _generateHasLongDesc(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return []
 
@@ -146,39 +219,155 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         return []
 
-    def _generateLabelOrName(self, obj, **args):
-        if self._script.utilities.isTextBlockElement(obj) \
-           and not self._script.utilities.isLandmark(obj) \
-           and not self._script.utilities.isDPub(obj):
+    def _generateHasDetails(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
-        if self._script.utilities.inDocumentContent(obj) and obj.name:
-            result = [obj.name]
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateHasDetails(obj, **args)
+
+        objs = self._script.utilities.detailsIn(obj)
+        if not objs:
+            return []
+
+        objString = lambda x: str.strip("%s %s" % (x.name, self.getLocalizedRoleName(x)))
+        toPresent = ", ".join(set(map(objString, objs)))
+
+        args['stringType'] = 'hasdetails'
+        result = [self._script.formatting.getString(**args) % toPresent]
+        result.extend(self.voice(speech_generator.SYSTEM))
+        return result
+
+    def _generateAllDetails(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        objs = self._script.utilities.detailsIn(obj)
+        if not objs:
+            container = pyatspi.findAncestor(obj, self._script.utilities.hasDetails)
+            objs = self._script.utilities.detailsIn(container)
+
+        if not objs:
+            return []
+
+        args['stringType'] = 'hasdetails'
+        result = [self._script.formatting.getString(**args) % ""]
+        result.extend(self.voice(speech_generator.SYSTEM))
+
+        result = []
+        for o in objs:
+            result.append(self.getLocalizedRoleName(o))
+            result.extend(self.voice(speech_generator.SYSTEM))
+
+            string = self._script.utilities.expandEOCs(o)
+            if not string.strip():
+                continue
+
+            result.append(string)
+            result.extend(self.voice(speech_generator.DEFAULT))
+            result.extend(self._generatePause(o))
+
+        return result
+
+    def _generateDetailsFor(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateDetailsFor(obj, **args)
+
+        objs = self._script.utilities.detailsFor(obj)
+        if not objs:
+            return []
+
+        if args.get('leaving'):
+            return []
+
+        lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        if (lastKey in ['Down', 'Right'] or self._script.inSayAll()) and args.get('startOffset'):
+            return []
+        if lastKey in ['Up', 'Left']:
+            text = self._script.utilities.queryNonEmptyText(obj)
+            if text and args.get('endOffset') not in [None, text.characterCount]:
+                return []
+
+        result = []
+        objArgs = {'stringType': 'detailsfor', 'mode': args.get('mode')}
+        for o in objs:
+            string = self._script.utilities.displayedText(o) or self.getLocalizedRoleName(o)
+            words = string.split()
+            if len(words) > 5:
+                words = words[0:5] + ['...']
+
+            result.append(self._script.formatting.getString(**objArgs) % " ".join(words))
+            result.extend(self.voice(speech_generator.SYSTEM))
+            result.extend(self._generatePause(o, **objArgs))
+
+        return result
+
+    def _generateLabelOrName(self, obj, **args):
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateLabelOrName(obj, **args)
+
+        if self._script.utilities.isTextBlockElement(obj) \
+           and not self._script.utilities.isLandmark(obj) \
+           and not self._script.utilities.isDPub(obj) \
+           and not self._script.utilities.isContentSuggestion(obj):
+            return []
+
+        if obj.name:
+            name = obj.name
+            if not self._script.utilities.hasExplicitName(obj):
+                name = name.strip()
+            result = [name]
             result.extend(self.voice(speech_generator.DEFAULT))
             return result
 
         return super()._generateLabelOrName(obj, **args)
 
     def _generateName(self, obj, **args):
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateName(obj, **args)
+
         if self._script.utilities.isTextBlockElement(obj) \
            and not self._script.utilities.isLandmark(obj) \
            and not self._script.utilities.isDPub(obj):
             return []
 
+        role = args.get('role', obj.getRole())
+        alwaysPresent = [pyatspi.ROLE_PUSH_BUTTON,
+                         pyatspi.ROLE_IMAGE]
+
+        if obj.parent and obj.name and obj.name == obj.parent.name \
+           and obj != orca_state.locusOfFocus and role not in alwaysPresent:
+            return []
+
         # TODO - JD: Once the formatting strings are vastly cleaned up
         # or simply removed, hacks like this won't be needed.
-        role = args.get('role', obj.getRole())
         if role in [pyatspi.ROLE_COMBO_BOX, pyatspi.ROLE_SPIN_BUTTON]:
             return super()._generateName(obj, **args)
 
-        if self._script.utilities.inDocumentContent(obj) and obj.name:
-            result = [obj.name]
+        if obj.name:
+            if self._script.utilities.preferDescriptionOverName(obj):
+                result = [obj.description]
+            elif self._script.utilities.isLink(obj) \
+                 and not self._script.utilities.hasExplicitName(obj):
+                return []
+            else:
+                name = obj.name
+                if not self._script.utilities.hasExplicitName(obj):
+                    name = name.strip()
+                result = [name]
+
             result.extend(self.voice(speech_generator.DEFAULT))
             return result
 
         return super()._generateName(obj, **args)
 
     def _generateLabel(self, obj, **args):
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateLabel(obj, **args)
+
         if self._script.utilities.isTextBlockElement(obj):
             return []
 
@@ -191,6 +380,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateLabel(obj, **args)
 
     def _generateNewNodeLevel(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if self._script.utilities.isTextBlockElement(obj) \
            or self._script.utilities.isLink(obj):
             return []
@@ -198,6 +390,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateNewNodeLevel(obj, **args)
 
     def _generateLeaving(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not args.get('leaving'):
             return []
 
@@ -216,7 +411,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return []
 
     def _generateNumberOfChildren(self, obj, **args):
-        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+        if _settingsManager.getSetting('onlySpeakDisplayedText') \
+           or _settingsManager.getSetting('speechVerbosityLevel') == settings.VERBOSITY_LEVEL_BRIEF:
             return []
 
         # We handle things even for non-document content due to issues in
@@ -259,6 +455,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return self._generateDisplayedText(obj, **args)
 
     def _generateRoleName(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateRoleName(obj, **args)
 
@@ -279,13 +478,17 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         force = args.get('force', False)
         start = args.get('startOffset')
         end = args.get('endOffset')
+        index = args.get('index', 0)
+        total = args.get('total', 1)
 
         if not force:
             doNotSpeak = [pyatspi.ROLE_FOOTER,
+                          pyatspi.ROLE_FORM,
                           pyatspi.ROLE_LABEL,
                           pyatspi.ROLE_MENU_ITEM,
                           pyatspi.ROLE_PARAGRAPH,
                           pyatspi.ROLE_SECTION,
+                          pyatspi.ROLE_REDUNDANT_OBJECT,
                           pyatspi.ROLE_UNKNOWN]
         else:
             doNotSpeak = [pyatspi.ROLE_UNKNOWN]
@@ -294,17 +497,26 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             doNotSpeak.append(pyatspi.ROLE_TABLE_CELL)
             doNotSpeak.append(pyatspi.ROLE_TEXT)
             doNotSpeak.append(pyatspi.ROLE_STATIC)
+            if args.get('string'):
+                doNotSpeak.append("ROLE_CONTENT_SUGGESTION")
             if args.get('formatType', 'unfocused') != 'basicWhereAmI':
                 doNotSpeak.append(pyatspi.ROLE_LIST_ITEM)
                 doNotSpeak.append(pyatspi.ROLE_LIST)
             if (start or end):
                 doNotSpeak.append(pyatspi.ROLE_DOCUMENT_FRAME)
+                doNotSpeak.append(pyatspi.ROLE_DOCUMENT_WEB)
                 doNotSpeak.append(pyatspi.ROLE_ALERT)
             if self._script.utilities.isAnchor(obj):
                 doNotSpeak.append(obj.getRole())
+            if total > 1:
+                doNotSpeak.append(pyatspi.ROLE_ROW_HEADER)
+            if self._script.utilities.isMenuInCollapsedSelectElement(obj):
+                doNotSpeak.append(pyatspi.ROLE_MENU)
 
-        if obj.getState().contains(pyatspi.STATE_EDITABLE):
-            lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        isEditable = obj.getState().contains(pyatspi.STATE_EDITABLE)
+
+        if isEditable and not self._script.utilities.isContentEditableWithEmbeddedObjects(obj):
             if ((lastKey in ["Down", "Right"] and not mods) or self._script.inSayAll()) and start:
                 return []
             if lastKey in ["Up", "Left"] and not mods:
@@ -315,16 +527,23 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 result.append(self.getLocalizedRoleName(obj, **args))
                 result.extend(acss)
 
+        elif isEditable and self._script.utilities.isDocument(obj):
+            if obj.parent and not obj.parent.getState().contains(pyatspi.STATE_EDITABLE) \
+               and lastKey not in ["Home", "End", "Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
+                result.append(object_properties.ROLE_EDITABLE_CONTENT)
+                result.extend(acss)
+
         elif role == pyatspi.ROLE_HEADING:
-            level = self._script.utilities.headingLevel(obj)
-            if level:
-                result.append(object_properties.ROLE_HEADING_LEVEL_SPEECH % {
-                    'role': self.getLocalizedRoleName(obj, **args),
-                    'level': level})
-                result.extend(acss)
-            else:
-                result.append(self.getLocalizedRoleName(obj, **args))
-                result.extend(acss)
+            if index == total - 1 or not self._script.utilities.isFocusableWithMathChild(obj):
+                level = self._script.utilities.headingLevel(obj)
+                if level:
+                    result.append(object_properties.ROLE_HEADING_LEVEL_SPEECH % {
+                        'role': self.getLocalizedRoleName(obj, **args),
+                        'level': level})
+                    result.extend(acss)
+                else:
+                    result.append(self.getLocalizedRoleName(obj, **args))
+                    result.extend(acss)
 
         elif self._script.utilities.isLink(obj):
             if obj.parent.getRole() == pyatspi.ROLE_IMAGE:
@@ -334,22 +553,22 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 if self._script.utilities.hasUselessCanvasDescendant(obj):
                     result.append(self.getLocalizedRoleName(obj, role=pyatspi.ROLE_IMAGE))
                     result.extend(acss)
-                result.append(self.getLocalizedRoleName(obj, **args))
-                result.extend(acss)
+                if index == total - 1 or not self._script.utilities.isFocusableWithMathChild(obj):
+                    result.append(self.getLocalizedRoleName(obj, **args))
+                    result.extend(acss)
 
         elif role not in doNotSpeak and args.get('priorObj') != obj:
             result.append(self.getLocalizedRoleName(obj, **args))
             result.extend(acss)
 
-        index = args.get('index', 0)
-        total = args.get('total', 1)
+        if self._script.utilities.isMath(obj) and not self._script.utilities.isMathTopLevel(obj):
+            return result
+
         ancestorRoles = [pyatspi.ROLE_HEADING, pyatspi.ROLE_LINK]
-        if index == total - 1 \
-           and (role == pyatspi.ROLE_IMAGE or self._script.utilities.queryNonEmptyText(obj)):
-            speakRoles = lambda x: x and x.getRole() in ancestorRoles
-            ancestor = pyatspi.findAncestor(obj, speakRoles)
-            if ancestor and ancestor.getRole() != role:
-                result.extend(self._generateRoleName(ancestor))
+        speakRoles = lambda x: x and x.getRole() in ancestorRoles
+        ancestor = pyatspi.findAncestor(obj, speakRoles)
+        if ancestor and ancestor.getRole() != role and (index == total - 1 or obj.name == ancestor.name):
+            result.extend(self._generateRoleName(ancestor))
 
         return result
 
@@ -422,10 +641,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
            and not _settingsManager.getSetting('enablePositionSpeaking'):
             return []
 
-        # TODO - JD: We cannot do this for XUL (or whatever Firefox is
-        # using in its non-webcontent dialogs)
-        #if not self._script.utilities.inDocumentContent(obj):
-        #    return super()._generatePositionInList(obj, **args)
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generatePositionInList(obj, **args)
+
         menuRoles = [pyatspi.ROLE_MENU_ITEM,
                      pyatspi.ROLE_TEAROFF_MENU_ITEM,
                      pyatspi.ROLE_CHECK_MENU_ITEM,
@@ -456,8 +674,30 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         result.extend(self.voice(speech_generator.SYSTEM))
         return result
 
-    def _generateTableCellRow(self, obj, **args):
+    def _generateRealTableCell(self, obj, **args):
+        result = super()._generateRealTableCell(obj, **args)
         if not self._script.inFocusMode():
+            return result
+
+        if _settingsManager.getSetting('speakCellCoordinates'):
+            label = self._script.utilities.labelForCellCoordinates(obj)
+            if label:
+                result.append(label)
+                result.extend(self.voice(speech_generator.SYSTEM))
+                return result
+
+            row, col = self._script.utilities.coordinatesForCell(obj)
+            if self._script.utilities.cellRowChanged(obj):
+                result.append(messages.TABLE_ROW % (row + 1))
+                result.extend(self.voice(speech_generator.SYSTEM))
+            if self._script.utilities.cellColumnChanged(obj):
+                result.append(messages.TABLE_COLUMN % (col + 1))
+                result.extend(self.voice(speech_generator.SYSTEM))
+
+        return result
+
+    def _generateTableCellRow(self, obj, **args):
+        if not self._script.utilities.inDocumentContent(obj):
             return super()._generateTableCellRow(obj, **args)
 
         if not self._script.utilities.shouldReadFullRow(obj):
@@ -465,11 +705,19 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         isRow = lambda x: x and x.getRole() == pyatspi.ROLE_TABLE_ROW
         row = pyatspi.findAncestor(obj, isRow)
-        if row and row.name:
+        if row and row.name and not self._script.utilities.isLayoutOnly(row):
             return self.generate(row)
 
         return super()._generateTableCellRow(obj, **args)
 
+    def _generateRowHeader(self, obj, **args):
+        # TODO - JD: _lastCommandWasCaretNav is private.
+        if self._script.utilities.lastInputEventWasLineNav() \
+           and self._script._lastCommandWasCaretNav:
+            return []
+
+        return super()._generateRowHeader(obj)
+
     def generateSpeech(self, obj, **args):
         if not self._script.utilities.inDocumentContent(obj):
             msg = "WEB: %s is not in document content. Calling default speech generator." % obj
@@ -500,9 +748,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         self._restoreRole(oldRole, args)
         msg = "WEB: Speech generation for document object %s complete:" % obj
-        debug.println(debug.LEVEL_INFO, msg)
+        debug.println(debug.LEVEL_INFO, msg, True)
         for element in result:
-            debug.println(debug.LEVEL_ALL, "           %s" % element)
+            debug.println(debug.LEVEL_ALL, "%s%s" % (' ' * 18, element))
 
         return result
 
@@ -511,7 +759,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             return []
 
         result = []
-        contents = self._script.utilities.filterContentsForPresentation(contents, False)
+        contents = self._script.utilities.filterContentsForPresentation(contents, True)
         msg = "WEB: Generating speech contents (length: %i)" % len(contents)
         debug.println(debug.LEVEL_INFO, msg, True)
         for i, content in enumerate(contents):
@@ -522,6 +770,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             utterance = self.generateSpeech(
                 obj, startOffset=start, endOffset=end, string=string,
                 index=i, total=len(contents), **args)
+            if isinstance(utterance, list):
+                isNotEmptyList = lambda x: not (isinstance(x, list) and not x)
+                utterance = list(filter(isNotEmptyList, utterance))
             if utterance and utterance[0]:
                 result.append(utterance)
                 args['priorObj'] = obj
